require "defines"

place_debug_marker = false

default_steps_per_tick = 5

belt_type =
{
    flat = 0,
    down = 1,
    up = 2,
}

to_ground_cost =
{
    none   =   2,
    medium =  10,
    high   = 100,
}

ground_move_cost =
{
    none   = 2,
    medium = 2,
    high   = 5,
}

corner_penalty = 2

function set_pos(tab, pos, key, val)
    if not tab then print("!tab, ", debug.traceback()) end
    if not pos then print("!pos, ", debug.traceback()) end
    tab[pos.x] = tab[pos.x] or {}
    tab[pos.x][pos.y] = tab[pos.x][pos.y] or {}
    tab[pos.x][pos.y][key] = val or true
end

--TODO this could be done better
function clear_pos(tab, pos, key)
    if not tab then print("!tab, ", debug.traceback()) end
    if not pos then print("!pos, ", debug.traceback()) end
    tab[pos.x] = tab[pos.x] or {}
    tab[pos.x][pos.y] = tab[pos.x][pos.y] or {}
    tab[pos.x][pos.y][key] = nil
end

function get_pos(tab, pos, key)
    if not tab then print("!tab, ", debug.traceback()) end
    if not pos then print("!pos, ", debug.traceback()) end
    return tab[pos.x] ~= nil and tab[pos.x][pos.y] ~= nil and tab[pos.x][pos.y][key]
end

function tile_is_passable(surface, pos)
    --this should cover everything we need
    return surface.can_place_entity{name="basic-transport-belt", position=pos}
end

function player_settings(player_index)
    if not global.beltplanner_settings then global.beltplanner_settings = {} end
    local s = global.beltplanner_settings
    if not s[player_index] then
        s[player_index] = {}
        local ps = s[player_index]
        ps.underground_avoidance = "medium"
        ps.underground_length = "short"
        ps.corner_penalty = true
        ps.steps_per_tick = default_steps_per_tick
    end
    return s[player_index]
end

function new_minheap()
    return {
        add = function(self, element, value)
            assert(not get_pos(self.pos, element.pos, element.belt_type))
            set_pos(self.pos, element.pos, element.belt_type)
            self.size = self.size + 1
            local cur_num = self.size
            self.heap[cur_num] = {element=element, value=value, heap_index=cur_num}
            self:percolate(cur_num)
        end,
        contains = function(self, element)
            return get_pos(self.pos, element.pos, element.belt_type)
        end,
        pop = function(self)
            assert(self.size > 0)
            local root = self.heap[1]
            assert(root.heap_index == 1)
            local min_element = root.element
            assert(get_pos(self.pos, min_element.pos, min_element.belt_type))
            clear_pos(self.pos, min_element.pos, min_element.belt_type)
            root.element = self.heap[self.size].element
            root.value = self.heap[self.size].value
            self.heap[self.size] = nil
            self.size = self.size - 1
            self:reheap(1)
            return min_element
        end,
        empty = function(self)
            return self.size == 0
        end,
        get = function(self, element)
            --FIXME ugh...
            for _, v in pairs(self.heap) do
                local a = v.element
                if a.pos.x == element.pos.x and a.pos.y == element.pos.y and a.belt_type == element.belt_type then
                    return v
                end
            end
            return nil
        end,
        reheap = function(self, node)
            local left = 2*node
            local right = left + 1
            local minpos = node
            if left <= self.size and self.heap[left].value < self.heap[minpos].value then
                minpos = left
            end
            if right <= self.size and self.heap[right].value < self.heap[minpos].value then
                minpos = right
            end
            if minpos ~= node then
                self:swap_node(minpos, node)
                self:reheap(minpos)
            end
        end,
        percolate = function(self, cur_num)
            local value = self.heap[cur_num].value
            while cur_num > 1 do
                local parent_num = math.floor(cur_num/2)
                assert(self.heap[cur_num].heap_index == cur_num)
                assert(self.heap[parent_num].heap_index == parent_num)
                if self.heap[parent_num].value <= value then
                    break
                else
                    self:swap_node(parent_num, cur_num)
                end
                cur_num = parent_num
            end
        end,
        swap_node = function(self, num_a, num_b)
            local a = self.heap[num_a]
            local b = self.heap[num_b]
            --heap_index stays the same
            local tmp = {element=a.element, value=a.value}
            a.element, a.value = b.element, b.value
            b.element, b.value = tmp.element, tmp.value
        end,
        pos = {},
        heap = {},
        size = 0,
    }
end

function get_dir(from, to)
    if from.x == to.x then
        assert(from.y ~= to.y)
        if (from.y > to.y) then return defines.direction.north end
        return defines.direction.south
    end
    assert(from.y == to.y)
    if from.x > to.x then return defines.direction.west end
    return defines.direction.east
end

function h(a, b)
    return math.abs(a.x-b.x) + math.abs(a.y-b.y)
end

function new_path_state(start, goal, player, type)
    local res = {
        start = {
            x = math.floor(start.x),
            y = math.floor(start.y),
        },
        goal = {
            x = math.floor(goal.x),
            y = math.floor(goal.y),
        },
        player = player,
        ps = player_settings(player.index),
        type = type,
        closed_set = {},
        open_set = new_minheap(),
        came_from = {},
        nodes_visited = 0,
    }
    res.open_set:add({pos=res.start, belt_type=belt_type.flat, g_score=0}, h(res.start, res.goal))
    if place_debug_marker then
        res.open_set.heap[1].element.marker = player.surface.create_entity{
                                                name="beltplanner-debug-open",
                                                position=res.start}
    end
    return res
end

function find_path(state)
    local function neighbors(elem)
        local function valid_neighbor(candidate_pos)
            return not get_pos(state.closed_set, candidate_pos, 1)
                and tile_is_passable(state.player.surface, candidate_pos)
        end
        local new_pos = {}

        if elem.belt_type == belt_type.flat then
            for _, dir in pairs({{0, -1}, {0, 1}, {-1, 0}, {1, 0}}) do
                local candidate_pos = { x = elem.pos.x+dir[1], y = elem.pos.y+dir[2] }
                local penalty = 0
                if state.ps.corner_penalty then
                    local prev = get_pos(state.came_from, elem.pos, elem.belt_type)
                    if prev then
                        if prev.pos.x - elem.pos.x ~= 0 then
                            --we're currently moving horizontally
                            if dir[1] ~= 0 then
                                --keep moving that way with no additional cost
                            else
                                penalty = corner_penalty
                            end
                        else
                            --moving vertically
                            if dir[2] ~= 0 then
                                --great!
                            else
                                penalty = corner_penalty
                            end
                        end
                    end
                end
                if valid_neighbor(candidate_pos) then
                    table.insert(new_pos, {pos=candidate_pos,
                                           belt_type=belt_type.flat,
                                           g_score=elem.g_score+1+penalty})
                    table.insert(new_pos, {pos=candidate_pos,
                                           belt_type=belt_type.down,
                                           dir=dir,
                                           g_score=elem.g_score
                                                  +to_ground_cost[state.ps.underground_avoidance]
                                                  +penalty})
                end
            end

        elseif elem.belt_type == belt_type.down then
            for i = 1, 5 do
                local candidate_pos = { x = elem.pos.x+elem.dir[1]*i,
                                        y = elem.pos.y+elem.dir[2]*i }
                if valid_neighbor(candidate_pos) then
                    local penalty
                    if state.ps.underground_length == "short" then
                        penalty = i * ground_move_cost[state.ps.underground_avoidance] - 1
                    else
                        penalty = (6 - i) * ground_move_cost[state.ps.underground_avoidance]
                    end
                    table.insert(new_pos, {pos=candidate_pos,
                                           belt_type=belt_type.up,
                                           dir=elem.dir,
                                           g_score=elem.g_score+penalty})
                end
            end

        else
            assert(elem.belt_type == belt_type.up)
            local candidate_pos = { x = elem.pos.x+elem.dir[1],
                                    y = elem.pos.y+elem.dir[2] }
            if valid_neighbor(candidate_pos) then
                table.insert(new_pos, {pos=candidate_pos,
                                       belt_type=belt_type.flat,
                                       g_score=elem.g_score+1})
                table.insert(new_pos, {pos=candidate_pos,
                                       belt_type=belt_type.down,
                                       dir=elem.dir,
                                       g_score=elem.g_score
                                              +to_ground_cost[state.ps.underground_avoidance]})
            end
        end

        return function(s, var) s.num=s.num+1 return s.tab[s.num-1] end,
            {num=1, tab=new_pos},
            new_pos[1]
    end

    local steps = 0
    while steps < state.ps.steps_per_tick and not state.open_set:empty() do
        steps = steps + 1
        local cur = state.open_set:pop() --this is the one with the lowest f_score

        if cur.pos.x == state.goal.x and cur.pos.y == state.goal.y
        and (cur.belt_type == belt_type.flat or cur.belt_type == belt_type.up) then
            state.path = {}
            local prev = get_pos(state.came_from, cur.pos, cur.belt_type)
            table.insert(state.path, {position=state.goal,
                                      direction=get_dir(prev.pos, cur.pos),
                                      belt_type=cur.belt_type})
            local next_pos = cur.pos
            cur = prev
            while cur do
                table.insert(state.path, {position=cur.pos,
                                          direction=get_dir(cur.pos, next_pos),
                                          belt_type=cur.belt_type})
                next_pos = cur.pos
                cur = get_pos(state.came_from, cur.pos, cur.belt_type)
            end
            return
        end

        set_pos(state.closed_set, cur.pos, 1)
        state.nodes_visited = state.nodes_visited + 1

        if place_debug_marker then
            assert(cur.marker.name == "beltplanner-debug-open")
            cur.marker.destroy()
            cur.marker = state.player.surface.create_entity{name="beltplanner-debug-closed",
                                                            position=cur.pos}
        end

        for new_pos in neighbors(cur) do
            if not state.open_set:contains(new_pos) then
                state.open_set:add(new_pos, new_pos.g_score + h(new_pos.pos, state.goal))
                set_pos(state.came_from, new_pos.pos, new_pos.belt_type, cur)
                if place_debug_marker then
                    new_pos.marker = state.player.surface.create_entity{name="beltplanner-debug-open",
                                                                        position=new_pos.pos}
                end
            else
                local existing_node = state.open_set:get(new_pos)
                if existing_node.element.g_score > new_pos.g_score then
                    if new_pos.marker then new_pos.marker.destroy() end
                    new_pos.marker = existing_node.element.marker
                    existing_node.element = new_pos
                    existing_node.value = new_pos.g_score + h(new_pos.pos, state.goal)
                    state.open_set:percolate(existing_node.heap_index)
                    set_pos(state.came_from, new_pos.pos, new_pos.belt_type, cur)
                end
            end
        end
    end
    return
end

function create_settings_gui(player)
    if player.gui.center.beltplanner_config then return end
    local ps = player_settings(player.index)
    local frame = player.gui.center.add{type="frame",
                                        name="beltplanner_config",
                                        direction="vertical",
                                        caption={"config-gui-caption"}}

    --need to stuff things into a flow even though the frame's direction already is vertical
    --otherwise multiple elements still get placed in the same row if they're small enough
    local config_flow = frame.add{type="flow",
                                  direction="vertical",
                                  name="beltplanner_config_flow"}

    config_flow.add{type="label",
                    name="beltplanner_label_underground_avoidance",
                    caption={"label-underground-avoidance"}}
    local avoidance_flow = config_flow.add{type="flow", name="beltplanner_avoidance_flow"}
    avoidance_flow.add{type="checkbox",
                       name="beltplanner_checkbox_avoidance_none",
                       state=(ps.underground_avoidance == "none"),
                       caption={"checkbox-avoidance-none"}}
    avoidance_flow.add{type="checkbox",
                       name="beltplanner_checkbox_avoidance_medium",
                       state=(ps.underground_avoidance == "medium"),
                       caption={"checkbox-avoidance-medium"}}
    avoidance_flow.add{type="checkbox",
                       name="beltplanner_checkbox_avoidance_high",
                       state=(ps.underground_avoidance == "high"),
                       caption={"checkbox-avoidance-high"}}

    config_flow.add{type="label",
                    name="beltplanner_label_length_pre",
                    caption={"label-length-pre"}}
    local length_flow = config_flow.add{type="flow", name="beltplanner_length_flow"}
    length_flow.add{type="checkbox",
                name="beltplanner_checkbox_short",
                state=(ps.underground_length == "short"),
                caption={"checkbox-length-short"}}
    length_flow.add{type="checkbox",
                name="beltplanner_checkbox_long",
                state=(ps.underground_length == "long"),
                caption={"checkbox-length-long"}}
    config_flow.add{type="label",
                    name="beltplanner_label_length_post",
                    caption={"label-length-post"}}

    config_flow.add{type="checkbox",
                    name="beltplanner_checkbox_corner_penalty",
                    state=ps.corner_penalty,
                    caption={"checkbox-corner-penalty"}}

    local steps_flow = config_flow.add{type="flow", name="beltplanner_steps_flow"}
    steps_flow.add{type="label",
                    name="beltplanner_config_label_steps",
                    caption={"label-steps-per-tick"}}
    steps_flow.add{type="textfield",
                    name="beltplanner_config_value_steps",
                    text=tostring(ps.steps_per_tick)}

    config_flow.add{type="button",
                    name="beltplanner_button_close",
                    caption={"button-close"}}
end

function create_state_gui(player)
    assert(not player.gui.top.beltplanner_state)
    local ps = player_settings(player.index)
    local frame = player.gui.left.add{type="frame",
                                      name="beltplanner_state",
                                      direction="vertical",
                                      caption={"state-gui-caption"}}
    local state_table = frame.add{type="table",
                                  name="beltplanner_state_table",
                                  colspan=2}

    state_table.add{type="label",
                      name="beltplanner_label_open_set",
                      caption="#open_set: "}
    state_table.add{type="label",
                      name="beltplanner_value_open_set",
                      caption="0"}

    state_table.add{type="label",
                        name="beltplanner_label_visited",
                        caption={"label-nodes-visited"}}
    state_table.add{type="label",
                        name="beltplanner_value_visited",
                        caption="0"}

    state_table.add{type="label",
                  name="beltplanner_label_cost",
                  caption={"label-cost"}}
    state_table.add{type="label",
                  name="beltplanner_value_cost",
                  caption="0"}

    state_table.add{type="label",
                    name="beltplanner_state_label_steps",
                    caption={"label-steps-per-tick"}}
    local steps_input = state_table.add{type="textfield",
                                        name="beltplanner_state_value_steps",
                                        text=tostring(ps.steps_per_tick)}
    steps_input.style.minimal_width = 50
    steps_input.style.maximal_width = 50

    state_table.add{type="button",
                    name="beltplanner_button_abort",
                    caption={"button-abort"}}
    state_table.add{type="button",
                    name="beltplanner_button_apply",
                    caption={"button-apply"}}
end

function update_state_gui(state)
    local ui = state.player.gui.left.beltplanner_state.beltplanner_state_table
    ui.beltplanner_value_open_set.caption = tostring(state.open_set.size)
    ui.beltplanner_value_visited.caption = tostring(state.nodes_visited)
    ui.beltplanner_value_cost.caption = tostring(state.open_set.heap[1].value)
end

script.on_init(function()
    --this also seems to run when the mod is added to an existing save
    global.beltplanner_start = {}
    global.beltplanner_state = {}
end)

function update_0_2_to_0_3()
    global.beltplanner_running = {}
    if global.beltplanner_settings then
        for _, ps in pairs(global.beltplanner_settings) do
            ps.steps_per_tick = default_steps_per_tick
        end
    end
    return "0.3.0"
end

update_from_version = {
    ["0.1.0"] = function()
        for _, force in pairs(game.forces) do
            if force.technologies["beltplanner-basic"].researched then
                force.recipes["beltplanner-config"].enabled = true
            end
        end
        return "0.2.0"
    end,
    ["0.2.0"] = update_0_2_to_0_3,
    ["0.2.1"] = update_0_2_to_0_3,
    ["0.3.0"] = function()
        global.beltplanner_running = nil
        global.beltplanner_state = {}
        return "0.3.1"
    end,
}

script.on_configuration_changed(function(event)
    --print(serpent.block(event))
    if event.mod_changes and event.mod_changes.beltplanner then
        local bp = event.mod_changes.beltplanner
        if not bp.old_version then return end
        local ver = bp.old_version
        while ver ~= "0.3.1" do
            ver = update_from_version[ver]()
        end
    end
end)

script.on_event(defines.events.on_built_entity, function(event)
    local e = event.created_entity
    local p = game.get_player(event.player_index)

    --FIXME this is getting awful
    local marker_type = string.match(e.name, "^beltplanner%-(.+)$")
    if marker_type ~= "basic" and marker_type ~= "fast"
    and marker_type ~= "express" and marker_type ~= "end"
    and marker_type ~= "config" then return end

    if global.beltplanner_state[event.player_index] then
        p.print({"path-finder-running"})
        p.insert{name=e.name}
        e.destroy()
        return
    end

    if marker_type == "config" then
        e.destroy()
        create_settings_gui(p)
        return
    end

    if marker_type == "end" then
        local start = global.beltplanner_start[event.player_index]
        if not start or not start.valid then
            p.print({"missing-start-marker"})
            e.destroy()
            return
        end

        local ui = p.gui.center.beltplanner_config
        if ui then
            update_steps_per_tick(player_settings(p.index),
                                  ui.beltplanner_config_flow
                                    .beltplanner_steps_flow
                                    .beltplanner_config_value_steps)
            p.insert{name="beltplanner-config"}
            ui.destroy()
        end

        local start_type = string.match(start.name, "^beltplanner%-(.+)$")
        assert(start_type == "basic" or start_type == "fast" or start_type == "express")
        assert(not global.beltplanner_state[event.player_index])
        global.beltplanner_state[event.player_index] = new_path_state(start.position, e.position, p, start_type)
        script.on_event(defines.events.on_tick, path_tick)
        start.destroy()
        global.beltplanner_start[event.player_index] = nil
        e.destroy()
        create_state_gui(p)
        return
    end

    --must have placed a start marker
    local prev_start = global.beltplanner_start[event.player_index]
    if prev_start and prev_start.valid then
        p.print({"multiple-start-markers"})
        p.insert{name=prev_start.name, count=1}
        prev_start.destroy()
    end
    global.beltplanner_start[event.player_index] = e
    local end_count = p.get_item_count("beltplanner-end")
    if end_count > 0 then
        p.remove_item{name="beltplanner-end", count=end_count}
    end
    if p.cursor_stack.valid_for_read and p.cursor_stack.count > 0 then
        assert(p.cursor_stack.name == "beltplanner-"..marker_type)
        local tmp_count = p.cursor_stack.count
        p.cursor_stack.clear()
        p.insert{name="beltplanner-"..marker_type, count=tmp_count}
    end
    p.cursor_stack.set_stack{name="beltplanner-end", count=1}
end)

function place_path(state)
    local name
    local type
    for _, v in pairs(state.path) do
        if v.belt_type == belt_type.flat then
            name = state.type.."-transport-belt"
            type = nil
        elseif v.belt_type == belt_type.down then
            name = state.type.."-transport-belt-to-ground"
            type = "input"
        else
            assert(v.belt_type == belt_type.up)
            name = state.type.."-transport-belt-to-ground"
            type = "output"
        end
        state.player.surface.create_entity{name="entity-ghost",
                                           inner_name=name,
                                           position=v.position,
                                           direction=v.direction,
                                           force=state.player.force,
                                           type=type}
    end
end

function path_tick(event)
    local removeme = {}

    for player_index, state in pairs(global.beltplanner_state) do
        find_path(state)
        if state.path then
            place_path(state)
            table.insert(removeme, player_index)
        elseif state.open_set:empty() then
            state.player.print({"no-path-found"})
            state.player.insert{name="beltplanner-"..state.type}
            table.insert(removeme, player_index)
        elseif event.tick % 60 == 0 then
            update_state_gui(state)
        end
    end

    for _, player_index in pairs(removeme) do
        global.beltplanner_state[player_index] = nil
        game.get_player(player_index).gui.left.beltplanner_state.destroy()
    end

    if not next(global.beltplanner_state) then
        script.on_event(defines.events.on_tick, nil)
    end
end

script.on_event(defines.events.on_player_mined_item, function(event)
    local item = event.item_stack
    assert(item.name ~= "beltplanner-end")

    local marker_type = string.match(item.name, "^beltplanner%-(.+)$")
    if marker_type ~= "basic" and marker_type ~= "fast"
    and marker_type ~= "express" and marker_type ~= "end" then return end

    local p = game.get_player(event.player_index)
    --this isn't necessarily the player who placed the start marker
    --but still, clearing any possible end markers should be the right thing to do
    local count = p.get_item_count("beltplanner-end")
    if count > 0 then
        p.remove_item{name="beltplanner-end", count=count}
    end
end)

function update_steps_per_tick(ps, steps_input)
    local steps = tonumber(steps_input.text)
    if not steps or steps < 1 then steps = 1
    elseif steps > 500 then steps = 500 end
    steps_input.text = steps
    ps.steps_per_tick = steps
end

script.on_event(defines.events.on_gui_click, function(event)
    local ps = player_settings(event.player_index)
    local ui = event.element.parent

    local name = event.element.name
    if name == "beltplanner_checkbox_avoidance_none" then
        ps.underground_avoidance = "none"
        ui.beltplanner_checkbox_avoidance_none.state = true --prevent this from getting unchecked by clicking again
        ui.beltplanner_checkbox_avoidance_medium.state = false
        ui.beltplanner_checkbox_avoidance_high.state = false
    elseif name == "beltplanner_checkbox_avoidance_medium" then
        ps.underground_avoidance = "medium"
        ui.beltplanner_checkbox_avoidance_none.state = false
        ui.beltplanner_checkbox_avoidance_medium.state = true
        ui.beltplanner_checkbox_avoidance_high.state = false
    elseif name == "beltplanner_checkbox_avoidance_high" then
        ps.underground_avoidance = "high"
        ui.beltplanner_checkbox_avoidance_none.state = false
        ui.beltplanner_checkbox_avoidance_medium.state = false
        ui.beltplanner_checkbox_avoidance_high.state = true
    elseif name == "beltplanner_checkbox_short" then
        ps.underground_length = "short"
        ui.beltplanner_checkbox_short.state = true
        ui.beltplanner_checkbox_long.state = false
    elseif name == "beltplanner_checkbox_long" then
        ps.underground_length = "long"
        ui.beltplanner_checkbox_short.state = false
        ui.beltplanner_checkbox_long.state = true
    elseif name == "beltplanner_checkbox_corner_penalty" then
        ps.corner_penalty = event.element.state
    elseif name == "beltplanner_button_close" then
        update_steps_per_tick(ps, ui.beltplanner_steps_flow.beltplanner_config_value_steps)
        ui.gui.player.insert{name="beltplanner-config"}
        ui.parent.destroy()
    elseif name == "beltplanner_button_abort" then
        local p = game.get_player(event.player_index)
        p.insert{name="beltplanner-"..global.beltplanner_state[event.player_index].type}
        global.beltplanner_state[event.player_index] = nil
        p.gui.left.beltplanner_state.destroy()
        if not next(global.beltplanner_state) then
            script.on_event(defines.events.on_tick, nil)
        end
    elseif name == "beltplanner_button_apply" then
        update_steps_per_tick(ps, ui.beltplanner_state_value_steps)
    end
end)

script.on_load(function()
    --on_load runs before on_configuration_changed, so need to deal with missing global.beltplanner_state
    if global.beltplanner_state and next(global.beltplanner_state) then
        script.on_event(defines.events.on_tick, path_tick)
    end
end)
